# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import opengeode_inspector as og_inspector
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeModelType

# Local application imports
from .geode_model import GeodeModel, ComponentRegistry


class GeodeSection(GeodeModel):
    section: og.Section

    def __init__(self, section: og.Section | None = None) -> None:
        self.section = section if section is not None else og.Section()
        super().__init__(self.section)

    @classmethod
    def geode_object_type(cls) -> GeodeModelType:
        return "Section"

    def native_extension(self) -> str:
        return self.section.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return False

    @classmethod
    def is_viewable(cls) -> bool:
        return True

    def builder(self) -> og.SectionBuilder:
        return og.SectionBuilder(self.section)

    @classmethod
    def load(cls, filename: str) -> GeodeSection:
        return GeodeSection(og.load_section(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.section_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_section_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.SectionInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.SectionOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.section_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_section_saveable(self.section, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_section(self.section, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_section(self.section, filename_without_extension)

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_section(
            self.section, filename_without_extension
        )

    def mesh_components(self) -> ComponentRegistry:
        return self.section.mesh_components()

    def collection_components(self) -> ComponentRegistry:
        return self.section.collection_components()

    def boundaries(self, id: og.uuid) -> list[og.ComponentID]:
        return self.section.boundaries(id)

    def internals(self, id: og.uuid) -> list[og.ComponentID]:
        return self.section.internals(id)

    def items(self, id: og.uuid) -> list[og.ComponentID]:
        return self.section.items(id)

    def inspect(self) -> og_inspector.SectionInspectionResult:
        return og_inspector.inspect_section(self.section)

    def assign_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.assign_section_geographic_coordinate_system_info(
            self.section, builder, crs_name, info
        )

    def convert_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.convert_section_coordinate_reference_system(
            self.section, builder, crs_name, info
        )

    def create_crs(
        self, crs_name: str, input: og.CoordinateSystem2D, output: og.CoordinateSystem2D
    ) -> None:
        builder = self.builder()
        og.create_section_coordinate_system(
            self.section, builder, crs_name, input, output
        )



"""Packages"""

import argparse
import os
from typing import Any
import flask
import flask_cors  # type: ignore
from flask import Flask, Response
from flask_cors import cross_origin
from werkzeug.exceptions import HTTPException
from opengeodeweb_back import utils_functions, app_config
from opengeodeweb_back.routes import blueprint_routes
from opengeodeweb_back.routes.create import blueprint_create
from opengeodeweb_microservice.database import connection


def create_app(name: str) -> flask.Flask:
    app = flask.Flask(name)

    """ Config variables """
    FLASK_DEBUG = (
        True if os.environ.get("FLASK_DEBUG", default=None) == "True" else False
    )
    if FLASK_DEBUG == False:
        app.config.from_object(app_config.ProdConfig)
        SECONDS_BETWEEN_SHUTDOWNS: float = float(
            app.config.get("SECONDS_BETWEEN_SHUTDOWNS") or 60.0
        )
        utils_functions.set_interval(
            utils_functions.kill_task, SECONDS_BETWEEN_SHUTDOWNS, app
        )
    else:
        app.config.from_object(app_config.DevConfig)

    @app.before_request
    def before_request() -> flask.Response | None:
        if flask.request.method == "OPTIONS":
            response = flask.make_response()
            response.headers["Access-Control-Allow-Methods"] = (
                "GET,POST,PUT,DELETE,OPTIONS"
            )
            return response
        utils_functions.before_request(flask.current_app)
        return None

    @app.teardown_request
    def teardown_request(exception: BaseException | None) -> None:
        utils_functions.teardown_request(flask.current_app, exception)

    @app.errorhandler(HTTPException)
    def errorhandler(exception: HTTPException) -> tuple[dict[str, Any], int] | Response:
        return utils_functions.handle_exception(exception)

    @app.errorhandler(Exception)
    def handle_generic_exception(exception: Exception) -> Response:
        print("\033[91mError:\033[0m \033[91m" + str(exception) + "\033[0m", flush=True)
        return flask.make_response({"description": str(exception)}, 500)

    @app.route(
        "/error",
        methods=["POST"],
    )
    def return_error() -> Response:
        flask.abort(500, f"Test")
        return flask.make_response({}, 500)

    @app.route("/", methods=["POST"])
    @cross_origin()
    def root() -> Response:
        return flask.make_response({}, 200)

    @app.route("/kill", methods=["POST"])
    @cross_origin()
    def kill() -> None:
        print("Manual server kill, shutting down...", flush=True)
        os._exit(0)

    return app


def register_ogw_back_blueprints(app: flask.Flask) -> None:
    app.register_blueprint(
        blueprint_routes.routes,
        url_prefix="/opengeodeweb_back",
        name="opengeodeweb_back",
    )
    app.register_blueprint(
        blueprint_create.routes,
        url_prefix="/opengeodeweb_back/create",
        name="opengeodeweb_create",
    )


def run_server(app: Flask) -> None:
    parser = argparse.ArgumentParser(
        prog="OpenGeodeWeb-Back", description="Backend server for OpenGeodeWeb"
    )
    parser.add_argument(
        "--host",
        type=str,
        default=app.config.get("DEFAULT_HOST"),
        help="Host to run on",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=app.config.get("DEFAULT_PORT"),
        help="Port to listen on",
    )
    parser.add_argument(
        "-d",
        "--debug",
        default=app.config.get("FLASK_DEBUG"),
        help="Whether to run in debug mode",
        action="store_true",
    )
    parser.add_argument(
        "-dfp",
        "--data_folder_path",
        type=str,
        default=app.config.get("DEFAULT_DATA_FOLDER_PATH"),
        help="Path to the folder where data is stored",
    )
    parser.add_argument(
        "-ufp",
        "--upload_folder_path",
        type=str,
        default=app.config.get("UPLOAD_FOLDER"),
        help="Path to the folder where uploads are stored",
    )
    parser.add_argument(
        "-origins",
        "--allowed_origins",
        default=app.config.get("ORIGINS"),
        help="Origins that are allowed to connect to the server",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        default=app.config.get("MINUTES_BEFORE_TIMEOUT"),
        help="Number of minutes before the server times out",
    )
    args = parser.parse_args()
    app.config.update(DATA_FOLDER_PATH=args.data_folder_path)
    app.config.update(
        EXTENSIONS_FOLDER_PATH=os.path.join(str(args.data_folder_path), "extensions")
    )
    app.config.update(UPLOAD_FOLDER=args.upload_folder_path)
    app.config.update(MINUTES_BEFORE_TIMEOUT=args.timeout)
    flask_cors.CORS(app, origins=args.allowed_origins)
    print(f"{args=}", flush=True)

    db_filename: str = app.config.get("DATABASE_FILENAME") or "project.db"
    db_path = os.path.join(str(args.data_folder_path), db_filename)
    os.makedirs(os.path.dirname(db_path), exist_ok=True)
    app.config["SQLALCHEMY_DATABASE_URI"] = f"sqlite:///{db_path}"
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    connection.init_database(db_path)
    print(f"Database initialized at: {db_path}", flush=True)
    app.run(
        debug=args.debug,
        host=args.host,
        port=args.port,
        ssl_context=app.config.get("SSL"),
    )


def run_opengeodeweb_back() -> None:
    app = create_app(__name__)
    register_ogw_back_blueprints(app)
    run_server(app)
    print("Server stopped", flush=True)


# ''' Main '''
if __name__ == "__main__":
    run_opengeodeweb_back()



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
from opengeodeweb_microservice.database.data_types import (
    GeodeMeshType,
    ViewerElementsType,
)

# Local application imports
from .geode_vertex_set import GeodeVertexSet


class GeodeGraph(GeodeVertexSet):
    graph: og.Graph

    def __init__(self, graph: og.Graph | None = None) -> None:
        self.graph = graph if graph is not None else og.Graph.create()
        super().__init__(self.graph)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "Graph"

    @classmethod
    def viewer_elements_type(cls) -> ViewerElementsType:
        return "edges"

    def native_extension(self) -> str:
        return self.graph.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return False

    @classmethod
    def is_viewable(cls) -> bool:
        return False

    def builder(self) -> og.GraphBuilder:
        return og.GraphBuilder.create(self.graph)

    @classmethod
    def load(cls, filename: str) -> GeodeGraph:
        return GeodeGraph(og.load_graph(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.graph_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_graph_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.GraphInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.GraphOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.graph_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_graph_saveable(self.graph, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_graph(self.graph, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return ""

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return ""

    def inspect(self) -> object:
        return None

    def edge_attribute_manager(self) -> og.AttributeManager:
        return self.graph.edge_attribute_manager()



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeModelType

# Local application imports
from .geode_cross_section import GeodeCrossSection


class GeodeImplicitCrossSection(GeodeCrossSection):
    implicit_cross_section: og_geosciences.ImplicitCrossSection

    def __init__(
        self, implicit_cross_section: og_geosciences.ImplicitCrossSection | None = None
    ) -> None:
        self.implicit_cross_section = (
            implicit_cross_section
            if implicit_cross_section is not None
            else og_geosciences.ImplicitCrossSection()
        )
        super().__init__(self.implicit_cross_section)

    @classmethod
    def geode_object_type(cls) -> GeodeModelType:
        return "ImplicitCrossSection"

    def native_extension(self) -> str:
        return self.implicit_cross_section.native_extension()

    def builder(self) -> og_geosciences.ImplicitCrossSectionBuilder:
        return og_geosciences.ImplicitCrossSectionBuilder(self.implicit_cross_section)

    @classmethod
    def load(cls, filename: str) -> GeodeImplicitCrossSection:
        return GeodeImplicitCrossSection(
            og_geosciences.load_implicit_cross_section(filename)
        )

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og_geosciences.implicit_cross_section_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og_geosciences.is_implicit_cross_section_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og_geosciences.ImplicitCrossSectionInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og_geosciences.ImplicitCrossSectionOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og_geosciences.implicit_cross_section_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og_geosciences.is_implicit_cross_section_saveable(
            self.implicit_cross_section, filename
        )

    def save(self, filename: str) -> list[str]:
        return og_geosciences.save_implicit_cross_section(
            self.implicit_cross_section, filename
        )

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_implicit_cross_section(
            self.implicit_cross_section, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_implicit_cross_section(
            self.implicit_cross_section, filename_without_extension
        )



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_inspector as og_inspector
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_surface_mesh2d import GeodeSurfaceMesh2D
from .geode_grid2d import GeodeGrid2D


class GeodeRegularGrid2D(GeodeSurfaceMesh2D, GeodeGrid2D):
    regular_grid: og.RegularGrid2D

    def __init__(self, regular_grid: og.RegularGrid2D | None = None) -> None:
        self.regular_grid = (
            regular_grid if regular_grid is not None else og.RegularGrid2D.create()
        )
        super().__init__(self.regular_grid)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "RegularGrid2D"

    def inspect(self) -> og_inspector.SurfaceInspectionResult:
        return super().inspect()

    def native_extension(self) -> str:
        return self.regular_grid.native_extension()

    def builder(self) -> og.RegularGridBuilder2D:
        return og.RegularGridBuilder2D.create(self.regular_grid)

    @classmethod
    def load(cls, filename: str) -> GeodeRegularGrid2D:
        return GeodeRegularGrid2D(og.load_regular_grid2D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.regular_grid_additional_files2D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_regular_grid_loadable2D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.RegularGridInputFactory2D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.RegularGridOutputFactory2D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.regular_grid_object_priority2D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_regular_grid_saveable2D(self.regular_grid, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_regular_grid2D(self.regular_grid, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_regular_grid2D(
            self.regular_grid, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_regular_grid2D(
            self.regular_grid, filename_without_extension
        )

    def cell_attribute_manager(self) -> og.AttributeManager:
        return self.regular_grid.cell_attribute_manager()



# Standard library imports
import os
import threading
import time
import xml.etree.ElementTree as ET
import zipfile
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from typing import Any

# Third party imports
import flask
import fastjsonschema  # type: ignore
import importlib.metadata as metadata
import shutil
from werkzeug.exceptions import HTTPException
import werkzeug
from opengeodeweb_microservice.schemas import SchemaDict
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session
from opengeodeweb_microservice.database.data_types import GeodeObjectType

# Local application imports
from . import geode_functions
from .geode_objects import geode_objects
from .geode_objects.geode_model import GeodeModel
from .geode_objects.geode_object import GeodeObject


def increment_request_counter(current_app: flask.Flask) -> None:
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
        REQUEST_COUNTER += 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def decrement_request_counter(current_app: flask.Flask) -> None:
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
        REQUEST_COUNTER -= 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def update_last_request_time(current_app: flask.Flask) -> None:
    if "LAST_REQUEST_TIME" in current_app.config:
        LAST_REQUEST_TIME = time.time()
        current_app.config.update(LAST_REQUEST_TIME=LAST_REQUEST_TIME)


def terminate_session(exception: BaseException | None) -> None:
    session = flask.g.pop("session", None)
    if session is None:
        return
    if exception is None:
        session.commit()
    else:
        session.rollback()
    session.close()


def before_request(current_app: flask.Flask) -> None:
    increment_request_counter(current_app)
    flask.g.session = get_session()
    flask.g.start_time = time.perf_counter()


def teardown_request(
    current_app: flask.Flask, exception: BaseException | None = None
) -> None:
    decrement_request_counter(current_app)
    update_last_request_time(current_app)
    terminate_session(exception)
    if flask.has_request_context():
        message = "Request to " + str(flask.request.endpoint) + " completed"
        if hasattr(flask.g, "start_time"):
            duration = time.perf_counter() - flask.g.start_time
            message += " in " + str(duration) + "s"
        print(message, flush=True)


def kill_task(current_app: flask.Flask) -> None:
    REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
    LAST_PING_TIME = float(current_app.config.get("LAST_PING_TIME", 0))
    LAST_REQUEST_TIME = float(current_app.config.get("LAST_REQUEST_TIME", 0))
    MINUTES_BEFORE_TIMEOUT = float(current_app.config.get("MINUTES_BEFORE_TIMEOUT", 0))
    current_time = time.time()
    minutes_since_last_request = (current_time - LAST_REQUEST_TIME) / 60
    minutes_since_last_ping = (current_time - LAST_PING_TIME) / 60

    if REQUEST_COUNTER > 0:
        return
    if MINUTES_BEFORE_TIMEOUT == 0:
        return
    if minutes_since_last_ping > MINUTES_BEFORE_TIMEOUT:
        kill_server()
    if minutes_since_last_request > MINUTES_BEFORE_TIMEOUT:
        kill_server()


def kill_server() -> None:
    print("Server timed out due to inactivity, shutting down...", flush=True)
    os._exit(0)


def versions(list_packages: list[str]) -> list[dict[str, str]]:
    list_with_versions = []
    for package in list_packages:
        list_with_versions.append(
            {"package": package, "version": metadata.distribution(package).version}
        )
    return list_with_versions


def validate_request(request: flask.Request, schema: SchemaDict) -> dict[str, Any]:
    json_data = request.get_json(force=True, silent=True)

    if json_data is None:
        json_data = {}
    try:
        validate = fastjsonschema.compile(schema)
        validate(json_data)
    except fastjsonschema.JsonSchemaException as e:
        error_msg = str(e)
        print("Validation failed:", error_msg, flush=True)
        flask.abort(400, error_msg)
    return json_data


def set_interval(
    function: Callable[[flask.Flask], None], seconds: float, args: flask.Flask
) -> threading.Timer:
    def function_wrapper() -> None:
        set_interval(function, seconds, args)
        function(args)

    timer = threading.Timer(seconds, function_wrapper)
    timer.daemon = True
    timer.start()
    return timer


def extension_from_filename(filename: str) -> str:
    return os.path.splitext(filename)[1][1:]


def send_file(
    upload_folder: str, saved_files: list[str], new_file_name: str
) -> flask.Response:
    if len(saved_files) == 1:
        mimetype = "application/octet-binary"
    else:
        mimetype = "application/zip"
        new_file_name = os.path.splitext(new_file_name)[0] + ".zip"
        with zipfile.ZipFile(
            os.path.join(os.path.abspath(upload_folder), new_file_name), "w"
        ) as zipObj:
            for saved_file_path in saved_files:
                zipObj.write(
                    saved_file_path,
                    os.path.basename(saved_file_path),
                )

    response = flask.send_from_directory(
        directory=os.path.abspath(upload_folder),
        path=new_file_name,
        as_attachment=True,
        mimetype=mimetype,
    )
    response.headers["new-file-name"] = new_file_name
    response.headers["Access-Control-Expose-Headers"] = "new-file-name"

    return response


def handle_exception(exception: HTTPException) -> flask.Response:
    print("\033[91mError:\033[0m \033[91m" + str(exception) + "\033[0m", flush=True)
    response = flask.jsonify(
        {
            "code": exception.code,
            "name": exception.name,
            "description": exception.description or "An error occurred",
        }
    )
    response.content_type = "application/json"
    response.status_code = exception.code or 500
    return response


def create_data_folder_from_id(data_id: str) -> str:
    base_data_folder = flask.current_app.config["DATA_FOLDER_PATH"]
    data_path = os.path.join(base_data_folder, data_id)
    os.makedirs(data_path, exist_ok=True)
    return data_path


def model_components(
    data_id: str, model: GeodeModel, viewable_file: str
) -> dict[str, Any]:
    vtm_file_path = geode_functions.data_file_path(data_id, viewable_file)
    tree = ET.parse(vtm_file_path)
    root = tree.find("vtkMultiBlockDataSet")
    if root is None:
        flask.abort(500, "Failed to read viewable file")
    uuid_to_flat_index = {}
    current_index = 0
    assert root is not None
    for elem in root.iter():
        if "uuid" in elem.attrib and elem.tag == "DataSet":
            uuid_to_flat_index[elem.attrib["uuid"]] = current_index
        current_index += 1
    model_mesh_components = model.mesh_components()
    mesh_components = []
    for mesh_component, ids in model_mesh_components.items():
        component_type = mesh_component.get()
        for id in ids:
            geode_id = id.string()
            component_name = geode_id
            viewer_id = uuid_to_flat_index[geode_id]
            boundaries = model.boundaries(id)
            boundaries_uuid = [boundary.id().string() for boundary in boundaries]
            internals = model.internals(id)
            internals_uuid = [internal.id().string() for internal in internals]
            mesh_component_object = {
                "viewer_id": viewer_id,
                "geode_id": geode_id,
                "name": component_name,
                "type": component_type,
                "boundaries": boundaries_uuid,
                "internals": internals_uuid,
            }
            mesh_components.append(mesh_component_object)

    model_collection_components = model.collection_components()
    collection_components = []
    for collection_component, ids in model_collection_components.items():
        component_type = collection_component.get()
        for id in ids:
            geode_id = id.string()
            items = model.items(id)
            items_uuid = [item.id().string() for item in items]
            collection_component_object = {
                "geode_id": geode_id,
                "name": geode_id,
                "type": component_type,
                "items": items_uuid,
            }
            collection_components.append(collection_component_object)
    return {
        "mesh_components": mesh_components,
        "collection_components": collection_components,
    }


def save_all_viewables_and_return_info(
    geode_object: GeodeObject,
    data: Data,
    data_path: str,
) -> dict[str, Any]:
    with ThreadPoolExecutor() as executor:
        native_files, viewable_path, light_path = executor.map(
            lambda args: args[0](args[1]),
            [
                (
                    geode_object.save,
                    os.path.join(
                        data_path, "native." + geode_object.native_extension()
                    ),
                ),
                (geode_object.save_viewable, os.path.join(data_path, "viewable")),
                (
                    geode_object.save_light_viewable,
                    os.path.join(data_path, "light_viewable"),
                ),
            ],
        )
        with open(light_path, "rb") as f:
            binary_light_viewable = f.read()
        data.native_file = os.path.basename(native_files[0])
        data.viewable_file = os.path.basename(viewable_path)
        data.light_viewable_file = os.path.basename(light_path)

        assert data.native_file is not None
        assert data.viewable_file is not None
        assert data.light_viewable_file is not None
        name = geode_object.identifier.name()
        if not name:
            flask.abort(400, "Geode object has no name defined.")

        response: dict[str, Any] = {
            "native_file": data.native_file,
            "viewable_file": data.viewable_file,
            "id": data.id,
            "name": name,
            "viewer_type": data.viewer_object,
            "binary_light_viewable": binary_light_viewable.decode("utf-8"),
            "geode_object_type": data.geode_object,
        }
        if isinstance(geode_object, GeodeModel):
            response |= model_components(data.id, geode_object, data.viewable_file)
        return response


def generate_native_viewable_and_light_viewable_from_object(
    geode_object: GeodeObject,
) -> dict[str, Any]:
    data = Data.create(
        geode_object=geode_object.geode_object_type(),
        viewer_object=geode_object.viewer_type(),
        viewer_elements_type=geode_object.viewer_elements_type(),
    )
    data_path = create_data_folder_from_id(data.id)
    return save_all_viewables_and_return_info(geode_object, data, data_path)


def generate_native_viewable_and_light_viewable_from_file(
    geode_object_type: GeodeObjectType, input_file: str
) -> dict[str, Any]:
    generic_geode_object = geode_objects[geode_object_type]
    data = Data.create(
        geode_object=geode_object_type,
        viewer_object=generic_geode_object.viewer_type(),
        viewer_elements_type=generic_geode_object.viewer_elements_type(),
    )
    data_path = create_data_folder_from_id(data.id)
    full_input_filename = geode_functions.upload_file_path(input_file)
    geode_object = generic_geode_object.load(full_input_filename)
    geode_object.builder().set_name(input_file)
    return save_all_viewables_and_return_info(geode_object, data, data_path)



# Standard library imports
import os
import time
import shutil
import math
from typing import Any

# Third party imports
import flask
import werkzeug
import zipfile
import opengeode as og
import opengeode_io as og_io
import opengeode_geosciences as og_geosciences
import opengeode_geosciencesio as og_geosciencesio
from opengeodeweb_microservice.schemas import get_schemas_dict
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session
from opengeodeweb_microservice.database import connection
from opengeodeweb_microservice.database.data_types import geode_object_type

# Local application imports
from opengeodeweb_back import geode_functions, utils_functions
from opengeodeweb_back.routes import schemas
from opengeodeweb_back.geode_objects import geode_objects
from opengeodeweb_back.geode_objects.geode_mesh import GeodeMesh
from opengeodeweb_back.geode_objects.geode_graph import GeodeGraph
from opengeodeweb_back.geode_objects.geode_grid2d import GeodeGrid2D
from opengeodeweb_back.geode_objects.geode_grid3d import GeodeGrid3D
from opengeodeweb_back.geode_objects.geode_surface_mesh2d import GeodeSurfaceMesh2D
from opengeodeweb_back.geode_objects.geode_surface_mesh3d import GeodeSurfaceMesh3D
from opengeodeweb_back.geode_objects.geode_solid_mesh3d import GeodeSolidMesh3D

routes = flask.Blueprint("routes", __name__, url_prefix="/opengeodeweb_back")

schemas_dict = get_schemas_dict(os.path.join(os.path.dirname(__file__), "schemas"))


@routes.route(
    schemas_dict["allowed_files"]["route"],
    methods=schemas_dict["allowed_files"]["methods"],
)
def allowed_files() -> flask.Response:
    utils_functions.validate_request(flask.request, schemas_dict["allowed_files"])
    extensions: set[str] = set()
    for geode_object in geode_objects.values():
        for extension in geode_object.input_extensions():
            extensions.add(extension)
    return flask.make_response({"extensions": list(extensions)}, 200)


@routes.route(
    schemas_dict["upload_file"]["route"],
    methods=schemas_dict["upload_file"]["methods"],
)
def upload_file() -> flask.Response:
    UPLOAD_FOLDER = flask.current_app.config["UPLOAD_FOLDER"]
    print(f"{UPLOAD_FOLDER=}", flush=True)
    if not os.path.exists(UPLOAD_FOLDER):
        os.makedirs(UPLOAD_FOLDER, exist_ok=True)

    file = flask.request.files["file"]
    if file.filename is None:
        flask.abort(400, "Filename is required")
    filename = werkzeug.utils.secure_filename(os.path.basename(file.filename))
    print(f"{filename=}", flush=True)
    file.save(os.path.join(UPLOAD_FOLDER, filename))
    return flask.make_response({"message": "File uploaded"}, 201)


@routes.route(
    schemas_dict["allowed_objects"]["route"],
    methods=schemas_dict["allowed_objects"]["methods"],
)
def allowed_objects() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["allowed_objects"]
    )
    params = schemas.AllowedObjects.from_dict(json_data)
    file_absolute_path = geode_functions.upload_file_path(params.filename)
    file_extension = utils_functions.extension_from_filename(
        os.path.basename(file_absolute_path)
    )
    allowed_objects = {}
    for geode_object_type, geode_object in geode_objects.items():
        if file_extension not in geode_object.input_extensions():
            continue
        loadability_score = geode_object.is_loadable(file_absolute_path)
        priority_score = geode_object.object_priority(file_absolute_path)
        allowed_objects[geode_object_type] = {
            "is_loadable": loadability_score.value(),
            "object_priority": priority_score,
        }
    return flask.make_response({"allowed_objects": allowed_objects}, 200)


@routes.route(
    schemas_dict["missing_files"]["route"],
    methods=schemas_dict["missing_files"]["methods"],
)
def missing_files() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["missing_files"]
    )
    params = schemas.MissingFiles.from_dict(json_data)
    file_path = geode_functions.upload_file_path(params.filename)
    geode_object = geode_functions.geode_object_from_string(params.geode_object_type)
    additional_files = geode_object.additional_files(
        file_path,
    )
    has_missing_files = any(
        file.is_missing
        for file in additional_files.mandatory_files + additional_files.optional_files
    )
    mandatory_files = [
        os.path.basename(file.filename)
        for file in additional_files.mandatory_files
        if file.is_missing
    ]
    additional_files_array = [
        os.path.basename(file.filename)
        for file in additional_files.optional_files
        if file.is_missing
    ]

    return flask.make_response(
        {
            "has_missing_files": has_missing_files,
            "mandatory_files": mandatory_files,
            "additional_files": additional_files_array,
        },
        200,
    )


@routes.route(
    schemas_dict["geographic_coordinate_systems"]["route"],
    methods=schemas_dict["geographic_coordinate_systems"]["methods"],
)
def crs_converter_geographic_coordinate_systems() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["geographic_coordinate_systems"]
    )
    params = schemas.GeographicCoordinateSystems.from_dict(json_data)
    geode_object = geode_functions.geode_object_from_string(params.geode_object_type)
    infos = (
        og_geosciences.GeographicCoordinateSystem3D.geographic_coordinate_systems()
        if geode_object.is_3D()
        else og_geosciences.GeographicCoordinateSystem2D.geographic_coordinate_systems()
    )
    crs_list = []
    for info in infos:
        crs = {}
        crs["name"] = info.name
        crs["code"] = info.code
        crs["authority"] = info.authority
        crs_list.append(crs)
    return flask.make_response({"crs_list": crs_list}, 200)


@routes.route(
    schemas_dict["inspect_file"]["route"],
    methods=schemas_dict["inspect_file"]["methods"],
)
def inspect_file() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["inspect_file"]
    )
    params = schemas.InspectFile.from_dict(json_data)
    file_path = geode_functions.upload_file_path(params.filename)
    geode_object = geode_functions.geode_object_from_string(
        params.geode_object_type
    ).load(file_path)
    inspection_data = geode_object.inspect()
    inspection_result = extract_inspector_result(inspection_data)
    return flask.make_response({"inspection_result": inspection_result}, 200)


def extract_inspector_result(inspection_data: Any) -> dict[str, Any]:
    new_object = {}

    if hasattr(inspection_data, "inspection_type"):
        new_object["title"] = inspection_data.inspection_type()
        new_object["nb_issues"] = 0
        new_object["children"] = []
        for child in dir(inspection_data):
            if child.startswith("__") or child in [
                "inspection_type",
                "string",
                "nb_issues",
            ]:
                continue
            child_instance = getattr(inspection_data, child)
            if callable(child_instance):
                continue
            child_object = extract_inspector_result(child_instance)
            new_object["children"].append(child_object)
            if "nb_issues" in child_object:
                new_object["nb_issues"] += child_object["nb_issues"]
    elif hasattr(inspection_data, "description"):
        new_object["title"] = inspection_data.description()
        nb_issues = inspection_data.nb_issues()
        new_object["nb_issues"] = nb_issues
        if nb_issues > 0:
            issues = inspection_data.string().split("\n")
            new_object["issues"] = issues
    return new_object


@routes.route(
    schemas_dict["geode_objects_and_output_extensions"]["route"],
    methods=schemas_dict["geode_objects_and_output_extensions"]["methods"],
)
def geode_objects_and_output_extensions() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["geode_objects_and_output_extensions"]
    )
    params = schemas.GeodeObjectsAndOutputExtensions.from_dict(json_data)
    file_path = geode_functions.upload_file_path(params.filename)
    geode_object = geode_functions.geode_object_from_string(
        params.geode_object_type
    ).load(file_path)
    geode_objects_and_output_extensions = (
        geode_functions.geode_object_output_extensions(geode_object)
    )
    return flask.make_response(
        {"geode_objects_and_output_extensions": geode_objects_and_output_extensions},
        200,
    )


@routes.route(
    schemas_dict["save_viewable_file"]["route"],
    methods=schemas_dict["save_viewable_file"]["methods"],
)
def save_viewable_file() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["save_viewable_file"]
    )
    params = schemas.SaveViewableFile.from_dict(json_data)
    return flask.make_response(
        utils_functions.generate_native_viewable_and_light_viewable_from_file(
            geode_object_type=geode_object_type(params.geode_object_type),
            input_file=params.filename,
        ),
        200,
    )


@routes.route(
    schemas_dict["texture_coordinates"]["route"],
    methods=schemas_dict["texture_coordinates"]["methods"],
)
def texture_coordinates() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["texture_coordinates"]
    )
    params = schemas.TextureCoordinates.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeSurfaceMesh2D | GeodeSurfaceMesh3D):
        flask.abort(400, f"{params.id} is not a GeodeSurfaceMesh")
    texture_coordinates = geode_object.texture_manager().texture_names()
    return flask.make_response({"texture_coordinates": texture_coordinates}, 200)


def attributes_metadata(manager: og.AttributeManager) -> list[dict[str, str | float]]:
    attributes: list[dict[str, str | float]] = []
    nb_elements = manager.nb_elements()
    for name in manager.attribute_names():
        attribute = manager.find_generic_attribute(name)
        min_value, max_value = -1.0, -1.0
        if attribute.is_genericable():
            values = []
            nb_items = attribute.nb_items()
            for index in range(nb_elements):
                for item in range(nb_items):
                    value = attribute.generic_item_value(index, item)
                    values.append(value)

            valid_values = [
                value for value in values if value is not None and not math.isnan(value)
            ]
            if valid_values:
                min_value, max_value = min(valid_values), max(valid_values)

        attributes.append(
            {"attribute_name": name, "min_value": min_value, "max_value": max_value}
        )
    return attributes


@routes.route(
    schemas_dict["vertex_attribute_names"]["route"],
    methods=schemas_dict["vertex_attribute_names"]["methods"],
)
def vertex_attribute_names() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["vertex_attribute_names"]
    )
    params = schemas.VertexAttributeNames.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeMesh):
        flask.abort(400, f"{params.id} is not a GeodeMesh")
    attribute_manager = geode_object.vertex_attribute_manager()
    return flask.make_response(
        {"attributes": attributes_metadata(attribute_manager)},
        200,
    )


@routes.route(
    schemas_dict["cell_attribute_names"]["route"],
    methods=schemas_dict["cell_attribute_names"]["methods"],
)
def cell_attribute_names() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["cell_attribute_names"]
    )
    params = schemas.CellAttributeNames.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeGrid2D | GeodeGrid3D):
        flask.abort(400, f"{params.id} is not a GeodeGrid")
    attribute_manager = geode_object.cell_attribute_manager()
    return flask.make_response(
        {"attributes": attributes_metadata(attribute_manager)},
        200,
    )


@routes.route(
    schemas_dict["polygon_attribute_names"]["route"],
    methods=schemas_dict["polygon_attribute_names"]["methods"],
)
def polygon_attribute_names() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["polygon_attribute_names"]
    )
    params = schemas.PolygonAttributeNames.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeSurfaceMesh2D | GeodeSurfaceMesh3D):
        flask.abort(400, f"{params.id} is not a GeodeSurfaceMesh")
    attribute_manager = geode_object.polygon_attribute_manager()
    return flask.make_response(
        {"attributes": attributes_metadata(attribute_manager)},
        200,
    )


@routes.route(
    schemas_dict["polyhedron_attribute_names"]["route"],
    methods=schemas_dict["polyhedron_attribute_names"]["methods"],
)
def polyhedron_attribute_names() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["polyhedron_attribute_names"]
    )
    params = schemas.PolyhedronAttributeNames.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeSolidMesh3D):
        flask.abort(400, f"{params.id} is not a GeodeSolidMesh")
    attribute_manager = geode_object.polyhedron_attribute_manager()
    return flask.make_response(
        {"attributes": attributes_metadata(attribute_manager)},
        200,
    )


@routes.route(
    schemas_dict["edge_attribute_names"]["route"],
    methods=schemas_dict["edge_attribute_names"]["methods"],
)
def edge_attribute_names() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["edge_attribute_names"]
    )
    params = schemas.EdgeAttributeNames.from_dict(json_data)
    geode_object = geode_functions.load_geode_object(params.id)
    if not isinstance(geode_object, GeodeGraph):
        flask.abort(400, f"{params.id} does not have edges")
    attribute_manager = geode_object.edge_attribute_manager()
    return flask.make_response(
        {"attributes": attributes_metadata(attribute_manager)},
        200,
    )


@routes.route(
    schemas_dict["ping"]["route"],
    methods=schemas_dict["ping"]["methods"],
)
def ping() -> flask.Response:
    utils_functions.validate_request(flask.request, schemas_dict["ping"])
    flask.current_app.config.update(LAST_PING_TIME=time.time())
    return flask.make_response({"message": "Flask server is running"}, 200)


@routes.route(schemas_dict["kill"]["route"], methods=schemas_dict["kill"]["methods"])
def kill() -> flask.Response:
    print("Manual server kill, shutting down...", flush=True)
    utils_functions.teardown_request(flask.current_app)
    os._exit(0)
    return flask.make_response({"message": "Flask server is dead"}, 200)


@routes.route(
    schemas_dict["export_project"]["route"],
    methods=schemas_dict["export_project"]["methods"],
)
def export_project() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["export_project"]
    )
    params = schemas.ExportProject.from_dict(json_data)

    project_folder: str = flask.current_app.config["DATA_FOLDER_PATH"]
    os.makedirs(project_folder, exist_ok=True)

    filename: str = werkzeug.utils.secure_filename(os.path.basename(params.filename))
    if not filename.lower().endswith(".vease"):
        flask.abort(400, "Requested filename must end with .vease")
    export_vease_path = os.path.join(project_folder, filename)

    with get_session() as session:
        rows = session.query(Data.id, Data.native_file).all()

    with zipfile.ZipFile(
        export_vease_path, "w", compression=zipfile.ZIP_DEFLATED
    ) as zip_file:
        database_root_path = os.path.join(project_folder, "project.db")
        if os.path.isfile(database_root_path):
            zip_file.write(database_root_path, "project.db")

        for data_id, native_file in rows:
            base_dir = os.path.join(project_folder, data_id)

            native_path = os.path.join(base_dir, str(native_file))
            if os.path.isfile(native_path):
                zip_file.write(native_path, os.path.join(data_id, str(native_file)))

        zip_file.writestr("snapshot.json", flask.json.dumps(params.snapshot))

    return utils_functions.send_file(project_folder, [export_vease_path], filename)


@routes.route(
    schemas_dict["import_project"]["route"],
    methods=schemas_dict["import_project"]["methods"],
)
def import_project() -> flask.Response:
    utils_functions.validate_request(flask.request, schemas_dict["import_project"])
    if "file" not in flask.request.files:
        flask.abort(400, "No .vease file provided under 'file'")
    zip_file = flask.request.files["file"]
    assert zip_file.filename is not None
    filename = werkzeug.utils.secure_filename(os.path.basename(zip_file.filename))
    if not filename.lower().endswith(".vease"):
        flask.abort(400, "Uploaded file must be a .vease")

    data_folder_path: str = flask.current_app.config["DATA_FOLDER_PATH"]

    # 423 Locked bypass : remove stopped requests
    if connection.scoped_session_registry:
        connection.scoped_session_registry.remove()
    if connection.engine:
        connection.engine.dispose()
    connection.engine = connection.session_factory = (
        connection.scoped_session_registry
    ) = None

    try:
        if os.path.exists(data_folder_path):
            shutil.rmtree(data_folder_path)
        os.makedirs(data_folder_path, exist_ok=True)
    except PermissionError:
        flask.abort(423, "Project files are locked; cannot overwrite")

    zip_file.stream.seek(0)
    with zipfile.ZipFile(zip_file.stream) as zip_archive:
        project_folder = os.path.abspath(data_folder_path)
        for member in zip_archive.namelist():
            target = os.path.abspath(
                os.path.normpath(os.path.join(project_folder, member))
            )
            if not (
                target == project_folder or target.startswith(project_folder + os.sep)
            ):
                flask.abort(400, "Vease file contains unsafe paths")
        zip_archive.extractall(project_folder)

        database_root_path = os.path.join(project_folder, "project.db")
        if not os.path.isfile(database_root_path):
            flask.abort(400, "Missing project.db at project root")

        connection.init_database(database_root_path, create_tables=False)

        try:
            with get_session() as session:
                rows = session.query(Data).all()
        except Exception:
            connection.init_database(database_root_path, create_tables=True)
            with get_session() as session:
                rows = session.query(Data).all()

        with get_session() as session:
            for data in rows:
                data_path = geode_functions.data_file_path(data.id)
                viewable_name = data.viewable_file
                if viewable_name:
                    vpath = geode_functions.data_file_path(data.id, viewable_name)
                    if os.path.isfile(vpath):
                        continue

                native_file = str(data.native_file or "")
                if not native_file:
                    continue

                native_full = geode_functions.data_file_path(data.id, native_file)
                if not os.path.isfile(native_full):
                    continue

                geode_object = geode_functions.geode_object_from_string(
                    data.geode_object
                ).load(native_full)
                utils_functions.save_all_viewables_and_return_info(
                    geode_object, data, data_path
                )
            session.commit()

        snapshot = {}
        try:
            raw = zip_archive.read("snapshot.json").decode("utf-8")
            snapshot = flask.json.loads(raw)
        except KeyError:
            snapshot = {}
    return flask.make_response({"snapshot": snapshot}, 200)


@routes.route(
    schemas_dict["geode_object_inheritance"]["route"],
    methods=schemas_dict["geode_object_inheritance"]["methods"],
)
def geode_object_inheritance() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["geode_object_inheritance"]
    )
    params = schemas.GeodeObjectInheritance.from_dict(json_data)
    geode_object_type = params.geode_object_type
    target_class = geode_functions.geode_object_from_string(geode_object_type)

    def get_all_bases(geode_class: type) -> set[type]:
        bases = set()
        for base_class in geode_class.__bases__:
            if base_class is not object:
                bases.add(base_class)
                bases.update(get_all_bases(base_class))
        return bases

    def get_all_subclasses(geode_class: type) -> set[type]:
        subclasses = set()
        for subclass_class in geode_class.__subclasses__():
            subclasses.add(subclass_class)
            subclasses.update(get_all_subclasses(subclass_class))
        return subclasses

    # Extract all related Geode classes (parents and children)
    base_classes = get_all_bases(target_class)
    subclass_classes = get_all_subclasses(target_class)

    # Filter GeodeObjectType to only include registered related objects, excluding target
    parents = []
    children = []
    for geode_object_type_str, geode_class in geode_objects.items():
        if geode_class == target_class:
            continue
        if geode_class in base_classes:
            parents.append(geode_object_type_str)
        if geode_class in subclass_classes:
            children.append(geode_object_type_str)

    return flask.make_response({"parents": parents, "children": children}, 200)



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import (
    GeodeMeshType,
    ViewerElementsType,
)

# Local application imports
from .geode_mesh import GeodeMesh


class GeodeRasterImage3D(GeodeMesh):
    raster_image: og.RasterImage3D

    def __init__(self, raster_image: og.RasterImage3D) -> None:
        self.raster_image = raster_image
        super().__init__(self.raster_image)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "RasterImage3D"

    @classmethod
    def viewer_elements_type(cls) -> ViewerElementsType:
        return "polyhedra"

    def native_extension(self) -> str:
        return self.raster_image.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return True

    @classmethod
    def is_viewable(cls) -> bool:
        return True

    def builder(self) -> None:
        return None

    @classmethod
    def load(cls, filename: str) -> GeodeRasterImage3D:
        return GeodeRasterImage3D(og.load_raster_image3D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.raster_image_additional_files3D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_raster_image_loadable3D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.RasterImageInputFactory3D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.RasterImageOutputFactory3D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.raster_image_object_priority3D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_raster_image_saveable3D(self.raster_image, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_raster_image3D(self.raster_image, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_raster_image3D(
            self.raster_image, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_raster_image3D(
            self.raster_image, filename_without_extension
        )

    def inspect(self) -> None:
        return None

    def vertex_attribute_manager(self) -> og.AttributeManager:
        return og.AttributeManager()



# Standard library imports
import os
import uuid

# Third party imports
from opengeodeweb_microservice.database.data_types import GeodeObjectType_values

# Local application imports
from opengeodeweb_back import geode_functions
from opengeodeweb_back.geode_objects import geode_objects

data_folder = os.path.join(os.path.dirname(__file__), "data")


def test_geode_objects() -> None:
    for geode_object_type in GeodeObjectType_values:
        assert geode_object_type in geode_objects


def test_input_output() -> None:
    for generic_geode_object in geode_objects.values():
        print(f"\n{generic_geode_object.geode_object_type()=}")
        for input_extension in generic_geode_object.input_extensions():
            print(f"\t{input_extension=}")
            file_absolute_path = os.path.join(data_folder, f"test.{input_extension}")
            if generic_geode_object.is_loadable(file_absolute_path).value() == 0.0:
                continue
            geode_object = generic_geode_object.load(file_absolute_path)
            data_name = geode_object.identifier.name() or os.path.basename(
                file_absolute_path
            )
            if geode_object.is_viewable():
                viewable_file_path = geode_object.save_viewable(
                    os.path.join(os.path.abspath(f"./output"), data_name)
                )
                os.remove(viewable_file_path)
            if geode_object.is_viewable():
                light_viewable_file_path = geode_object.save_light_viewable(
                    os.path.join(os.path.abspath(f"./output"), data_name)
                )
                os.remove(light_viewable_file_path)
            geode_objects_output_extensions = (
                geode_functions.geode_object_output_extensions(geode_object)
            )
            assert type(geode_objects_output_extensions) is dict
            for (
                output_geode_object_type,
                output_geode_extensions,
            ) in geode_objects_output_extensions.items():
                print(f"\t\t{output_geode_object_type=}")
                for (
                    output_extension,
                    output_is_saveable,
                ) in output_geode_extensions.items():
                    print(f"\t\t\t{output_extension=}")
                    uu_id = str(uuid.uuid4()).replace("-", "")
                    filename = f"{uu_id}.{output_extension}"
                    if output_is_saveable:
                        saved_files = geode_objects[output_geode_object_type].save(
                            geode_object,
                            os.path.join(os.path.abspath(f"./output"), filename),
                        )
                        assert type(saved_files) is list
                        for saved_file in saved_files:
                            os.remove(saved_file)



from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass


@dataclass
class Point(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    name: str
    x: float
    y: float
    z: float



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeModelType

# Local application imports
from .geode_section import GeodeSection


class GeodeCrossSection(GeodeSection):
    cross_section: og_geosciences.CrossSection

    def __init__(
        self, cross_section: og_geosciences.CrossSection | None = None
    ) -> None:
        self.cross_section = (
            cross_section
            if cross_section is not None
            else og_geosciences.CrossSection()
        )
        super().__init__(self.cross_section)

    @classmethod
    def geode_object_type(cls) -> GeodeModelType:
        return "CrossSection"

    def native_extension(self) -> str:
        return self.cross_section.native_extension()

    def builder(self) -> og_geosciences.CrossSectionBuilder:
        return og_geosciences.CrossSectionBuilder(self.cross_section)

    @classmethod
    def load(cls, filename: str) -> GeodeCrossSection:
        return GeodeCrossSection(og_geosciences.load_cross_section(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og_geosciences.cross_section_additional_files(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og_geosciences.is_cross_section_loadable(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og_geosciences.CrossSectionInputFactory.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og_geosciences.CrossSectionOutputFactory.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og_geosciences.cross_section_object_priority(filename)

    def is_saveable(self, filename: str) -> bool:
        return og_geosciences.is_cross_section_saveable(self.cross_section, filename)

    def save(self, filename: str) -> list[str]:
        return og_geosciences.save_cross_section(self.cross_section, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_cross_section(
            self.cross_section, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_cross_section(
            self.cross_section, filename_without_extension
        )



# Standard library imports
from __future__ import annotations
from abc import abstractmethod

# Third party imports
import opengeode as og
from opengeodeweb_microservice.database.data_types import ViewerElementsType

# Local application imports
from .geode_mesh import GeodeMesh


class GeodeGrid3D(GeodeMesh):
    @classmethod
    def is_3D(cls) -> bool:
        return True

    @classmethod
    def is_viewable(cls) -> bool:
        return True

    @classmethod
    def viewer_elements_type(cls) -> ViewerElementsType:
        return "polyhedra"

    def builder(self) -> object:
        return None

    def inspect(self) -> object:
        return None

    @abstractmethod
    def cell_attribute_manager(self) -> og.AttributeManager: ...



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import opengeode_geosciences as og_geosciences
import opengeode_inspector as og_inspector
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_graph import GeodeGraph


class GeodeEdgedCurve3D(GeodeGraph):
    edged_curve: og.EdgedCurve3D

    def __init__(self, edged_curve: og.EdgedCurve3D | None = None) -> None:
        self.edged_curve = (
            edged_curve if edged_curve is not None else og.EdgedCurve3D.create()
        )
        super().__init__(self.edged_curve)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "EdgedCurve3D"

    def native_extension(self) -> str:
        return self.edged_curve.native_extension()

    @classmethod
    def is_3D(cls) -> bool:
        return True

    @classmethod
    def is_viewable(cls) -> bool:
        return True

    def builder(self) -> og.EdgedCurveBuilder3D:
        return og.EdgedCurveBuilder3D.create(self.edged_curve)

    @classmethod
    def load(cls, filename: str) -> GeodeEdgedCurve3D:
        return GeodeEdgedCurve3D(og.load_edged_curve3D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.edged_curve_additional_files3D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_edged_curve_loadable3D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.EdgedCurveInputFactory3D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.EdgedCurveOutputFactory3D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.edged_curve_object_priority3D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_edged_curve_saveable3D(self.edged_curve, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_edged_curve3D(self.edged_curve, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_edged_curve3D(
            self.edged_curve, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_edged_curve3D(
            self.edged_curve, filename_without_extension
        )

    def inspect(self) -> og_inspector.EdgedCurveInspectionResult:
        return og_inspector.inspect_edged_curve3D(self.edged_curve)

    def assign_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.assign_edged_curve_geographic_coordinate_system_info3D(
            self.edged_curve, builder, crs_name, info
        )

    def convert_crs(
        self, crs_name: str, info: og_geosciences.GeographicCoordinateSystemInfo
    ) -> None:
        builder = self.builder()
        og_geosciences.convert_edged_curve_coordinate_reference_system3D(
            self.edged_curve, builder, crs_name, info
        )

    def create_crs(
        self, crs_name: str, input: og.CoordinateSystem2D, output: og.CoordinateSystem2D
    ) -> None:
        builder = self.builder()
        og.create_edged_curve_coordinate_system3D(
            self.edged_curve, builder, crs_name, input, output
        )



# Standard library imports
from __future__ import annotations

# Third party imports
import opengeode as og
import geode_viewables as viewables
from opengeodeweb_microservice.database.data_types import GeodeMeshType

# Local application imports
from .geode_solid_mesh3d import GeodeSolidMesh3D


class GeodeTetrahedralSolid3D(GeodeSolidMesh3D):
    tetrahedral_solid: og.TetrahedralSolid3D

    def __init__(self, tetrahedral_solid: og.TetrahedralSolid3D | None = None) -> None:
        self.tetrahedral_solid = (
            tetrahedral_solid
            if tetrahedral_solid is not None
            else og.TetrahedralSolid3D.create()
        )
        super().__init__(self.tetrahedral_solid)

    @classmethod
    def geode_object_type(cls) -> GeodeMeshType:
        return "TetrahedralSolid3D"

    def native_extension(self) -> str:
        return self.tetrahedral_solid.native_extension()

    def builder(self) -> og.TetrahedralSolidBuilder3D:
        return og.TetrahedralSolidBuilder3D.create(self.tetrahedral_solid)

    @classmethod
    def load(cls, filename: str) -> GeodeTetrahedralSolid3D:
        return GeodeTetrahedralSolid3D(og.load_tetrahedral_solid3D(filename))

    @classmethod
    def additional_files(cls, filename: str) -> og.AdditionalFiles:
        return og.tetrahedral_solid_additional_files3D(filename)

    @classmethod
    def is_loadable(cls, filename: str) -> og.Percentage:
        return og.is_tetrahedral_solid_loadable3D(filename)

    @classmethod
    def input_extensions(cls) -> list[str]:
        return og.TetrahedralSolidInputFactory3D.list_creators()

    @classmethod
    def output_extensions(cls) -> list[str]:
        return og.TetrahedralSolidOutputFactory3D.list_creators()

    @classmethod
    def object_priority(cls, filename: str) -> int:
        return og.tetrahedral_solid_object_priority3D(filename)

    def is_saveable(self, filename: str) -> bool:
        return og.is_tetrahedral_solid_saveable3D(self.tetrahedral_solid, filename)

    def save(self, filename: str) -> list[str]:
        return og.save_tetrahedral_solid3D(self.tetrahedral_solid, filename)

    def save_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_viewable_tetrahedral_solid3D(
            self.tetrahedral_solid, filename_without_extension
        )

    def save_light_viewable(self, filename_without_extension: str) -> str:
        return viewables.save_light_viewable_tetrahedral_solid3D(
            self.tetrahedral_solid, filename_without_extension
        )
